/* gencode.c
	code generator som-c		2 Jan 2004
	som3   						4 Nov 2004
*/

#include "compile.h"

#define	PRIVATE	static
#define	IM		(MAXCS - MAXDS - 1000)  /* immediate CS */

extern int CP;				/* current code pointer */

PRIVATE int currentf;		/* current function idx */

PRIVATE int item2(int x){ return car(cdr(x)); }
PRIVATE int item3(int x){ return car(cdr(cdr(x))); }

PRIVATE int atomeq(int e, int type, int value){
	return (car(e) == type) && (cdr(e) == value);
}

// e = (lv (+ lv 1))
PRIVATE int isinc(int e){
	int a, ref, op;
	a = car(e);
	if( car(a) != LNAME ) return 0;
	ref = cdr(a);
	e = item2(e);		// e = (+ lv 1)
	op = car(e);
	if( !(atomeq(op,OPER,tkPLUS) || atomeq(op,OPER,tkMINUS)) )
		return 0;
	a = item2(e);
	if( ! atomeq(a,LNAME,ref) ) return 0;
	if( ! atomeq(item3(e),NUM,1) ) return 0;
	return 1;
}

PRIVATE void genbop(int op, int l){
	genex(car(l));
	genex(item2(l));
	outs(op);
}

PRIVATE void genuop(int op, int l){
	genex(car(l));
	outs(op);
}

// gen vec address:  (name ex) => ld/get name, ex
PRIVATE void genvec(int e){
	int nm, type, ref;
	nm = car(e);
	type = car(nm);
	ref = cdr(nm);
	switch(type){
	case GNAME: outa(icLd,getRef(ref)); break;
	case LNAME: outa(icGet,ref); break;
	}
	genex(item2(e));
}

// return patch-address  e = (e1 e)
PRIVATE int genif(int e){
	int ads;
	genex(car(e));
	outa(icJf,0);
	ads = CP-1;
	genex(item2(e));
	return ads;
}

// if there is any break, patch jump to CP
void patchbreak(void){
	int ads;
	if( ytos() == MARK ){	// TO without break
		ypop();
		return;
	}
	while( ytos() == BMARK ){
		ypop();			// throw away BMARK
		ads = ypop();
		patch(ads,CP-ads);
	}
	if( ytos() != MARK )
		seterror("patch-break no matching MARK");
	ypop();				// throw away MARK
}

// e = (block ex ... ex), ex = (num ex)
// get lo and hi labels from caselist
PRIVATE void getcaselist(int e, int *lo, int *hi){
	int n, ex;
	*lo = 0;
	*hi = 0;
	if( e == NIL ) return;
	e = cdr(e);
	ex = car(e);  		// a = (num ex)
	n = cdr(car(ex));  	// n = num.ref
	*lo = n;
	*hi = n;
	while( e != NIL ){
		ex = car(e);
		n = cdr(car(ex));
		if( n < 0 ) break;  // else case
		if( n < *lo ) *lo = n;
		if( n > *hi ) *hi = n;
		e = cdr(e);
	}
}
// high quality, e = (ex0 block)
PRIVATE void gencase(int e){
	int lo, hi, ads, ads2, i, v1, v2, a;
	int type, ref;

	genex(car(e));				// gen ex0
	outs(icCase);
	e = item2(e);				// block
	getcaselist(e,&lo,&hi);
	outa(icLit,lo);
	outa(icLit,hi);
	ads = CP;
	outa(icJmp,0);				// else case
	for(i=lo; i<=hi; i++)
		outa(icJmp,0);			// empty jump table
	v1 = 0;						// v1 is backchained
	e = cdr(e);					// e = (ex .. ex)
	while( e != NIL ){
		a = car(e);				// a = (num ex)
		ref = cdr(car(a)); 		// ref = num label
		if( ref > 0 ){			// skip else case ref -1
			ads2 = ref-lo+ads+1;// entry in jump table
			patch(ads2,CP-ads2);// to jump here
			genex(item2(a));
			outa(icJmp,v1);		// jump to eloc, backchain
			v1 = CP-1;
		}
		e = cdr(e);
	}
	// here is exelse
	// patch the empty entry in jumptable to exelse
	for(i=ads+1; i<=ads+1+hi-lo; i++){
		decode(i,&type,&v2);
		if(v2 == 0) patch(i,CP-i);
	}
	patch(ads,CP-ads);	// the jmp to exelse
	if( ref == -1 )		// gen else case
		genex(item2(a));
	// here is eloc
	// update backchain jmp to eloc
	while( v1 != 0 ){	// update jmp to eloc
		decode(v1,&type,&v2);
		patch(v1,CP-v1);
		v1 = v2;
	}
	patch(v1,CP-v1);	// the first one
}
// (ex..ex)
PRIVATE void genlist(int e){
	while( e != NIL ){
		genex(car(e));
		e = cdr(e);
	}
}

void genex(int x){
	int a, e, e1, ads, ads2, v1, v2;
	int idx, type, ref;
	int i, lo, hi;

	if( x == NIL ) return;
	if(isatom(x)){
		type = car(x);
		ref = cdr(x);
		switch(type){
		case STRING:
		case NUM: outa(icLit,ref); break;
		case GNAME: outa(icLd,getRef(ref)); break;
		case LNAME: outa(icGet,ref); break;
		case OPER:
			switch(ref){
			case tkBREAK:
				ypush(CP);
				ypush(BMARK);
				outa(icJmp,0);
				break;
			}
			break;
		}
	}else{
		a = car(x);			// (op ex .. )
		e = cdr(x);			// e = (ex .. )
		if(car(a) != OPER)
			seterror("genex: expect operator");
		switch(cdr(a)){
		case tkPLUS: genbop(icAdd,e); break;
		case tkMINUS: genbop(icSub,e); break;
		case tkSTAR: genbop(icMul,e); break;
		case tkSLASH: genbop(icDiv,e); break;
		case tkAND: genbop(icBand,e); break;
		case tkBAR: genbop(icBor,e); break;
		case tkCARET: genbop(icBxor,e); break;
		case tkEQEQ: genbop(icEq,e); break;
		case tkNE: genbop(icNe,e); break;
		case tkLT: genbop(icLt,e); break;
		case tkLE: genbop(icLe,e); break;
		case tkGE: genbop(icGe,e); break;
		case tkGT: genbop(icGt,e); break;
		case tkMOD: genbop(icMod,e); break;
		case tkGTGT: genbop(icShr,e); break;
		case tkLTLT: genbop(icShl,e); break;
		case tkNOT: genuop(icNot,e); break;
		case tkARRAY: genuop(icArray,e); break;
		case tkSYSCALL: 			// e = (num ex0 .. ex0)
			genlist(cdr(e));
			outa(icSys,cdr(car(e)));
			break;
		case tkBB:					// (block ex .. ex)
			genlist(e);				// e = (ex..ex)
			break;
		case tkBE:					// (call name ex ... ex)
			idx = cdr(car(e));		// e = (name ex .. ex)
			genlist(cdr(e));
			a = getRef(idx);
			if( a == 0 )
				outa(icCalli,idx);
			else
				outa(icCall,getRef(idx));
			break;
		case tkTO:					// (to name ex)
			idx = cdr(car(e));		// e = (name ex)
			currentf = idx;
			e = item2(e);
			if( e != NIL ){
				setRef(idx,CP);
				outa(icFun,0);		/* clean will update it */
				ypush(MARK);
				genex(e);
				clean(idx);
			}
			break;
		case tkEQ:				// (= var/vec ex)
			a = car(e);			// e = (var/vec ex)
			e1 = item2(e);		// ex
			if(isinc(e)){		// e = (lv (+ lv 1))
				ref = cdr(a); 	// lvar ref
				if(atomeq(car(e1),OPER,tkPLUS))  // e1 = (+ lv 1)
					outa(icInc,ref);
				else
					outa(icDec,ref);
			}else if(isatom(a)){// var
				genex(e1);		// RHS
				type = car(a);
				ref = cdr(a);
				switch(type){
				case GNAME: outa(icSt,getRef(ref)); break;
				case LNAME: outa(icPut,ref); break;
				}
			}else{				// vec
				genvec(cdr(a));	// vec LHS
				genex(e1);		// RHS
				outs(icStx);
			}
			break;
		case tkLBRACKET:		// vec RHS
			genvec(e);
			outs(icLdx);
			break;
		case tkIF:				// (if e1 e)
			ads = genif(e);		// e = (e1 e)
			patch(ads,CP-ads);
			break;
		case tkELSE: 			// (ifelse e1 et ef)
			ads = genif(e);		// e = (e1 et ef)
			outa(icJmp,0);
			patch(ads,CP-ads);
			ads = CP-1;
			genex(item3(e));
			patch(ads,CP-ads);
			break;
		case tkWHILE: 			// (while e1 e)
			ypush(MARK);
			ads = CP;			// e = (e1 e)
			outa(icJmp,0);		// branch to cond
			genex(item2(e));	// body of while
			ads2 = CP;
			genex(car(e));		// cond
			outa(icJt,ads+1-CP);// jump to body
			patch(ads,ads2-ads);
			patchbreak();
			break;
		case tkFOR:				// (for lv ex0 ex0 ex)
			ypush(MARK);		// e = (lv ex0 ex0 ex)
			v2 = getLv(currentf) + 1;
			setLv(currentf,v2);	// update symtab
			v1 = cdr(car(e));
			e = cdr(e);			// e = (ex0 ex0 ex)
			genex(car(e));		// initial
			outa(icPut,v1);
			genex(item2(e));	// end
			outa(icPut,v2);
			ads = CP;
			outa(icJmp,0);		// jump to test
			genex(item3(e));	// body of for
			outa(icInc,v1);
			patch(ads,CP-ads);	// test
    		outa(icGet,v1);
			outa(icGet,v2);
			outs(icLe);
			outa(icJt,ads+1-CP);
			patchbreak();
			break;
		case tkCASE:			// e = (ex0 block)
			gencase(e);
			break;
		}
	}
}

// e = (block ex ex ... )
void genimm(int e){
	int idx, d, mainref;
	patchCalli(1,CP-1);
	idx = search("_main");
	mainref = CP;
	setRef(idx,mainref);
	outa(icFun,1);	// func _main
	e = cdr(e);
	while( e != NIL ){
		genex(car(e));
		e = cdr(e);
	}
	outa(icRet,1);
    d = CP;
	CP = 1;
	outa(icCall,mainref);
	outs(icEnd);
	CP = d;			// restore CP
}

// execute immediate line, use CS[IM] area
void runimm(int e){
	int savecp;
	savecp = CP;
	CP = IM;
	genex(e);
	outs(icEnd);
	CP = savecp;
	eval(IM);
}
